//! Implementation of [W3C RDF tests](http://w3c.github.io/rdf-tests/) to tests Rio parsers conformance.
#![deny(unsafe_code)]

mod isomorphism;
pub mod manifest;
pub mod model;
pub mod parser_evaluator;
pub mod report;
mod vocab;
